'   Program:    ANSI Slave Protocol Driver Software
'   Module:     ANSI_COMM.BAS
'   Platform:   MC204, MC216 to be checked
'-------------------------------------------------------------------
'   Version:    1.3
'   Author:     Gareth Jones
'   Modified:   09/04/98 GJ
'-------------------------------------------------------------------
'   Copyright 1999 Control Techniques
'
'   Control techniques
'   Software & System Products Division (SSPD)
'   Unit 79, Mochdre Ind. Est.
'   Newtown
'   Powys
'   SY16 4LE
'   gjones_sspd@compuserve.com
'   01686 612411
'====================================================================
'Description
'   Control Techniques use an ANSI Protocol to communicate between a master and
'a drive, this program is designed to mimic a drive to allow the MC Range of
'controllers to be interfaced to a system, and also to allow the use of the
'low cost hi spec CTIU MMI.  This intial version is designed only to respond to
'requests from a master connected to port 2 of the MC204 fitted with the RS485
'option, or Port 2 of the MC216 connected through a RS232 to RS485 external
'converter.  Later vesions may have the ability to function as an ANSI master,
'enquire for details.
'
'The protocol is as follows:
'
'Node Address:
'Each drive or controller on the system has a unique address made up of two
'digits, Group address, and Unit address, the addresses each have a range 1-9.
'Addresses 01, 02, 10, 20, 30 ect are invalid.
'Special ASCII charachters are used to frame the data these are as follows:
'
'Char       ASCII
'----------------
'STX        02
'ETX        03
'EOT        04
'ENQ        05
'ACK        06
'Backspace  08
'NAK        21
'
'Read data from the MC Controller
'--------------------------------
'
'the message to the MC Controller is assembled as follows:
'    |  Address  | Parameter |
'|eot|ga|ga|ua|ua|m1|m2|p1|p2|enq|
'
'ga = group address
'ua = unit address
'm  = menu
'p  = parameter
'
'The controller should reply with:
'    | Parameter |   Data  |
'|stx|m1|m2|p1|p2|d1|...|dn|etx|bcc|
'
'd  = data
'bcc= block checksum
'm  = menu
'p  = paramete
'
'Write data to an MC Controller
'------------------------------
'
'the message to the MC Controller is assembled as follows:
'
'    |  Address  |   | parameter |  data   |
'|eot|ga|ga|ua|ua|stx|m1|m2|p1|p2|d1|...|dn|etx|bcc|
'
'ga = group address
'ua = unit address
'm  = menu
'p  = parameter
'd  = data
'bcc= block checksum
'
'The drive will respond with a single control charachter:
'ACK or NAK
'
'If the address from the master is 00 then all the nodes listen.
'If the address from the master is 10, 20, 30, ... then all the nodes
'which start with the group address will listen ie. 1X, 2X, 3X etc.
'in these modes none of the nodes return an ACK or NAK.
'===============================================================================
'
'===============================================================================
'Set Up - Setup the port and data parameters
'===============================================================================
'
init_var:
'Set ANSI protocol parameters
'Set Port
    port=1

'Set the group address
    ga = 1

'Set the unit address
    ua = 1


' The table is split up into four areas integer, 1 decimal place, 2 decimal
' place and 3 decimal places, these areas can be allocated as required
' but must not overlap, and should be sequential.

' INTEGER +/-32000
    int_lo = 500
    int_hi = 2000
' 1 DECIMAL PLACE +/- 3200.0
    dec1_lo = 2500
    dec1_hi = 2999
' 2 DECIMAL PLACES +/- 320.00
    dec2_lo = 3000
    dec2_hi = 3499
' 3 DECIMAL PLACES +/- 32.000
    dec3_lo = 3500
    dec3_hi = 4999

' Part of table is used to build charachter strings, this will occupy about 20
' table values, this area should not overlap any of the other table values
    char_table = 2000

'Set comms port as 19200 baud, 7 data bits, 1 stop bit, even parity
    SETCOM(38400,7,1,2,port)

'set TABLE values
    GOSUB set_tab


'===============================================================================
'ansi - read and write using ansi protocol MAIN ROUTEIN
'-------------------------------------------------------------------------------
ansi:
'Get the address from the serial port
    GOSUB address_data
    IF addrs = FALSE THEN GOTO ansi
'Check if the request is a read or write command
    GOSUB read_write
'Get the parameter from the serial port
    GOSUB getparam
'Check that the menu & parameter data are within the valid ranges
    IF read = 1 THEN
        GET #port,gv_key
    ENDIF
    IF param_ok = FALSE THEN
        GOSUB nak
        GOTO ansi
    ENDIF

'Set the Variable type
    IF param >= int_lo AND param <= int_hi THEN param_type = 0
    IF param >= dec1_lo AND param <= dec1_hi THEN param_type = 1
    IF param >= dec2_lo AND param <= dec2_hi THEN param_type = 2
    IF param >= dec3_lo AND param <= dec3_hi THEN param_type = 3

'If the request is a read request then send the data
    IF read = 1 THEN GOSUB send_param
'If the request is a write request then get the data from the serial port
    IF write= 1 THEN
        GOSUB getval
        IF gv_retstat = FALSE THEN
            GOSUB nak
            GOTO ansi
        ENDIF
        GOSUB bcc
        GOSUB bcc_comp
        IF bcc_ok = TRUE THEN ' Write value to table
            TABLE (param,gv_retval)
            GOSUB ack
        ELSE
            GOSUB nak
        ENDIF
    ENDIF
GOTO ansi
'-------------------------------------------------------------------------------

'===============================================================================
'address_data - Decode Address
'-------------------------------------------------------------------------------
address_data:
        addrs = TRUE
    GET #port,gv_key

'look for control charachters
    IF gv_key <> 4 THEN GOTO address_fault
'-------------------------------------------------------------------------------
    GET #port,gv_key
    gv_key=TABLE(gv_key)

'look for group address
    IF (gv_key <> ga) AND (gv_key <> 0) THEN GOTO address_fault

'check if broadcast message is being sent
    IF gv_key = 0 THEN
        broadcast = TRUE
    ELSE
        broadcast = FALSE
    ENDIF
    GET #port,gv_key
    gv_key=TABLE(gv_key)

'look for group address 2nd transmission
    IF (gv_key <> ga) AND (gv_key <> 0) THEN
        GOSUB nak
        GOTO address_fault
    ENDIF
'-------------------------------------------------------------------------------
    GET #port,gv_key
    gv_key=TABLE(gv_key)

'look for unit address
    IF (gv_key <> ua) AND (gv_key <> 0) THEN GOTO address_fault

'check if broadcast message is being sent
    IF gv_key = 0 THEN
        broadcast = TRUE
    ELSE
        broadcast = FALSE
    ENDIF
    GET #port,gv_key
    gv_key=TABLE(gv_key)

'look for unit address 2nd transmission
    IF (gv_key <> ua) AND (gv_key <> 0) THEN
        GOSUB nak
        GOSUB address_fault
    ENDIF
RETURN
'Check if faults
address_fault:
    addrs = FALSE
RETURN

'===============================================================================
'read_write - Decode type of command, read or write
'-------------------------------------------------------------------------------
read_write:
'-------------------------------------------------------------------------------
    GET #port,gv_key
    TABLE ((char_table + gv_digits),gv_key)

'Check if read or write command by looking for control charachter STX
'Load Menu & Parameter Data
    IF gv_key <> 2 THEN
            read = 1
            write = 0
    ELSE
        write = 1
        read = 0
        GET #port,gv_key
        gv_digits = gv_digits + 1
        TABLE ((char_table + gv_digits),gv_key)
    ENDIF
RETURN


'===============================================================================
'===============================================================================
'                               Recieve data
'===============================================================================
'===============================================================================





'===============================================================================
'get_param - get the parameter data from the serial port
'-------------------------------------------------------------------------------
' getparam
'-------------------------------------------------------------------------------


getparam:
' NOTE: A GET statement has already got the first digit from the serial port
    gv_sign=1
    gv_int=0
    gv_dec=0
    gv_div=10
    gv_count=0
    gv_digits=0
    IF gv_key>57 OR gv_key<48 THEN
        GOTO gvdone
    ENDIF
    gv_digits = gv_digits + 1
    TABLE ((char_table + gv_digits),gv_key)
    gv_key=TABLE(gv_key)
i1:
    gv_count=gv_count+1
    gv_int=gv_int*10+gv_key

integer:
    IF gv_count = 4 THEN
        GOTO gvdone
    ENDIF
    GET #port,gv_key
    gv_digits = gv_digits + 1
    TABLE ((char_table + gv_digits),gv_key)
    IF gv_key>57 OR gv_key<48 THEN
        GOTO gvdone
    ENDIF
    gv_key=TABLE(gv_key)
    GOTO i1

gvdone:
' Check that four digits have been recieved for the parameter value
' and it is within range
    IF gv_count<>4 OR gv_int > dec3_hi OR gv_int < int_lo THEN
        param_ok=FALSE
        param = 0
    ELSE
        param_ok=TRUE
    ENDIF
    param = gv_int
RETURN



'===============================================================================
'getval - Get a decimal number from the serial port
'-------------------------------------------------------------------------------
getval:
'-------------------------------------------------------------------------------
    gv_sign=1
    gv_int=0
    gv_dec=0
    gv_div=10
    gv_count=0

    GET #port,gv_key
    gv_digits = gv_digits + 1
    TABLE ((char_table + gv_digits),gv_key)
    IF gv_key>57 OR gv_key<45 AND gv_key <> 43 THEN
        GOTO data_done
    ENDIF
    gv_key=TABLE(gv_key)
' Decimal point?
checkdp:
    IF gv_key=46 THEN
        GOTO data_dec
    ENDIF
' Minus?
    IF gv_key = 45 THEN
        gv_sign=-1
        GOTO data_int
    ENDIF
' Plus?
    IF gv_key = 43 THEN
        gv_sign= 1
        GOTO data_int
    ENDIF
    GOTO data_i1
data_int:
    GET #port,gv_key
    gv_digits = gv_digits + 1
    TABLE ((char_table + gv_digits),gv_key)
    IF gv_key>57 OR gv_key<46 OR gv_key=47 THEN
        GOTO data_done
    ENDIF
    gv_key=TABLE(gv_key)
data_i1:
    IF gv_key=46 THEN
        GOTO data_dec
    ENDIF
    gv_count=gv_count+1
    gv_int=gv_int*10+gv_key
    GOTO data_int

data_dec:
    GET #port,gv_key
    gv_digits = gv_digits + 1
    TABLE ((char_table + gv_digits),gv_key)
    IF gv_key>57 OR gv_key<48 THEN
        GOTO data_done
    ENDIF
    gv_key=TABLE(gv_key)
    gv_count=gv_count+1
    gv_dec=gv_dec+(gv_key/gv_div)
    gv_div=gv_div*10
    GOTO data_dec
data_done:
    IF gv_count=0 THEN
        gv_retstat=FALSE
    ELSE
        gv_retstat=TRUE
    ENDIF
    gv_retval=(gv_int+gv_dec)*gv_sign
    gv_retchar=gv_key
RETURN

'===============================================================================
'bcc - block CHECKSUM calculator FOR recieved bcc
'-------------------------------------------------------------------------------
bcc:
'-------------------------------------------------------------------------------
'Block checksum for recieved information
    bcc_check = 0
    FOR count = 1 TO gv_digits
        bcc_check = bcc_check XOR TABLE((char_table + count))
    NEXT count
    IF bcc_check < 32 THEN bcc_check = bcc_check + 32
RETURN
bcc_comp:
    GET #port,gv_key
    IF gv_key = bcc_check THEN
        bcc_ok = TRUE
    ELSE
        bcc_ok = FALSE
    ENDIF
RETURN



'===============================================================================
'===============================================================================
'                               transmit data
'===============================================================================
'===============================================================================




'===============================================================================
'nak - Send NAK
'-------------------------------------------------------------------------------
nak:
'-------------------------------------------------------------------------------
'Send NAK (Not Acknowledge)
    IF broadcast = TRUE THEN RETURN
    PRINT #port, CHR(21);
RETURN



'===============================================================================
'ack - Send ACK
'-------------------------------------------------------------------------------
ack:
'-------------------------------------------------------------------------------
'Send ACK (Acknowledge)
    IF broadcast = TRUE THEN RETURN
    PRINT #port, CHR(6);
RETURN



'===============================================================================
'Send_param - Send Parameter data
'-------------------------------------------------------------------------------
send_param:
'-------------------------------------------------------------------------------
'Reply to a read command
'Send STX
    PRINT #port, CHR(2);
'Send Menu & Parameter
    FOR count = 1 TO 4
        PRINT #port, (TABLE(char_table+count)-48)[0];
    NEXT count
'Send Data
    data = TABLE(param)
    IF TABLE(param) < 0 THEN
        PRINT #port, CHR(45);
        data = data * -1
        TABLE((char_table + 5), 45)
    ELSE
        PRINT #port, CHR(43);
        TABLE((char_table + 5), 43)
    ENDIF
    int_val = data

'The data needs to contain a fixed number of integer digits, this calculates
'the zeros put in front of the data so that it is the correct length
'and also checks that the data is within range
    IF param_type = 0 THEN
        IF int_val > 32767 THEN int_val = 32767
        IF int_val < -32768 THEN int_val = -32768
        int_data = int_val
    ENDIF

    IF param_type = 1 THEN
        IF int_val > 3276.7 THEN int_val = 3276.7
        IF int_val < -3276.8 THEN int_val = -3276.8
        int_data = int_val * 10
    ENDIF

    IF param_type = 2 THEN
        IF int_val > 327.67 THEN int_val = 327.67
        IF int_val < -327.68 THEN int_val = -327.68
        int_data = int_val * 100
    ENDIF

    IF param_type = 3 THEN
        IF int_val > 32.767 THEN int_val = 32.767
        IF int_val < -32.768 THEN int_val = -32.768
        int_data = int_val * 1000
    ENDIF

     gv_digits = 5
send_data:
    IF int_data < 10 THEN
        PRINT #port, "0000";
        GOTO send_last
    ENDIF
    FOR a = 4 TO 1 STEP -1
        gv_digits = gv_digits + 1
        mult_val = 1
        FOR b = 1 TO a
            mult_val = mult_val * 10
        NEXT b
        digit = INT(int_data / mult_val)
        'PRINT #6, digit;
        'round_var = digit
        'GOSUB round
        'digit = round_var
        int_data = int_data - (mult_val * digit)
        round_var = int_data
        GOSUB round
        int_data = round_var
        PRINT #port, digit[0];
        TABLE((char_table + gv_digits), (digit + 48))
' Print decimal point
        IF a = (param_type) THEN
            gv_digits = gv_digits + 1
            PRINT #port, CHR(46);
            TABLE((char_table + gv_digits), 46)
        ENDIF
    NEXT a
send_last:
        gv_digits = gv_digits + 1
        digit = int_data
        'PRINT #6, digit
        GOSUB round
        TABLE((char_table + gv_digits), (digit + 48))
        PRINT #port, digit[0];

end_message:
    gv_digits = gv_digits + 1
    TABLE((char_table + gv_digits), 3)
    PRINT #port, CHR(3); ' Send ETX
    GOSUB bcc
    PRINT #port, CHR(bcc_check); ' Send BCC
RETURN


'===============================================================================
'set_tab - table entries to define the numbers as 0 to 9.
'-------------------------------------------------------------------------------
set_tab:
    TABLE(40,40,41,42,43,44)
    TABLE(45,45,46,47)
    TABLE(48,00,01,02,03,04)
    TABLE(53,05,06,07,08,09)
RETURN


round:
    digit_int = INT(round_var)
    digit_fra = FRAC(round_var)
    IF digit_fra > 0.9 THEN digit_int = (digit_int + 1)
    round_var = digit_int
RETURN
